import * as si from "./DSI";
import { XMLDate, XMLElement, XMLNumber, XMLBoolean, XMLList, INamespaced, ISchemaInformation, IAttributes } from "./BaseTypes";
import { ensureArray } from "./Util";

const schemaVersion = "3.2.1";
const namespace = "dcc";
const namespaceUrl = `https://ptb.de/${namespace}`;
const schemaLocation = `${namespaceUrl}/v${schemaVersion}/dcc.xsd`;

export const DCC: ISchemaInformation = {
  schemaVersion,
  namespace,
  namespaceUrl,
  schemaLocation,
};

export class DCCXMLElement extends XMLElement implements INamespaced {
  get ns() {
    return DCC.namespace;
  }
}

export class DCCXMLBoolean extends XMLBoolean implements INamespaced {
  get ns() {
    return DCC.namespace;
  }
}

export class DCCXMLNumber extends XMLNumber implements INamespaced {
  get ns() {
    return DCC.namespace;
  }
}

export class DCCXMLDate extends XMLDate implements INamespaced {
  get ns() {
    return DCC.namespace;
  }
}

export class DCCXMLList extends XMLList implements INamespaced {
  get ns() {
    return DCC.namespace;
  }
}

export interface IIdAttributes extends IAttributes {
  id?: string;
}

export interface IIdAndRefTypeAttributes extends IAttributes {
  id?: string;
  refType?: string;
}

export interface IIdAndRefIdAndRefTypeAttributes extends IAttributes {
  id?: string;
  refId?: string;
  refType?: string;
}

export interface IDigitalCalibrationCertificateAttributes extends IAttributes {
  "xmlns:xsi"?: string;
  "xsi:schemaLocation"?: string;
  "xmlns:dcc"?: string;
  "xmlns:si"?: string;
  schemaVersion?: string;
}

/** The root element that contains the four rings of the DCC. */
export class DigitalCalibrationCertificateType extends DCCXMLElement {
  _attr: IDigitalCalibrationCertificateAttributes;

  administrativeData: AdministrativeDataType;
  measurementResults: MeasurementResultListType;
  comment?: Comment;
  document?: ByteDataType;
  "ds:Signature"?: never;

  constructor(el: Partial<DigitalCalibrationCertificateType> = {}) {
    super(el);
    this.administrativeData = new AdministrativeDataType(el.administrativeData);
    this.measurementResults = new MeasurementResultListType(el.measurementResults);
    if (el.comment) this.comment = new Comment(el.comment);
    if (el.document) this.document = new ByteDataType(el.document);
    if (el["ds:Signature"]) this["ds:Signature"] = el["ds:Signature"];
  }
}

/** The element administrativeData contains all essential administrative information about the calibration.
 * The entries in this area are basically the same and regulated in all DCCs. */
export class AdministrativeDataType extends DCCXMLElement {
  dccSoftware: SoftwareListType;
  refTypeDefinitions?: RefTypeDefinitionListType;
  coreData: CoreDataType;
  items: ItemListType;
  calibrationLaboratory: CalibrationLaboratoryType;
  respPersons: RespPersonListType;
  customer: ContactType;
  statements?: StatementListType;

  constructor(el: Partial<AdministrativeDataType> = {}) {
    super(el);
    this.dccSoftware = new SoftwareListType(el.dccSoftware);
    if (el.refTypeDefinitions) this.refTypeDefinitions = new RefTypeDefinitionListType(el.refTypeDefinitions);
    this.coreData = new CoreDataType(el.coreData);
    this.items = new ItemListType(el.items);
    this.calibrationLaboratory = new CalibrationLaboratoryType(el.calibrationLaboratory);
    this.respPersons = new RespPersonListType(el.respPersons);
    this.customer = new ContactType(el.customer);
    if (el.statements) this.statements = new StatementListType(el.statements);
  }
}

/* This type allows the give information about the used refTypes in a DCC. */
export class RefTypeDefinitionListType extends DCCXMLElement {
  refTypeDefinition: RefTypeDefinitionType[];

  constructor(el: Partial<RefTypeDefinitionListType> = {}) {
    super(el);
    this.refTypeDefinition = ensureArray(el.refTypeDefinition).map((x) => new RefTypeDefinitionType(x));
  }
}

/* This type contains the information about the wording of the refTypes used in a DCC. */
export class RefTypeDefinitionType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  name: TextType;
  description?: RichContentType;
  namespace: DCCXMLElement;
  link: DCCXMLElement;
  release?: DCCXMLElement;
  value?: DCCXMLElement;
  procedure?: DCCXMLElement;

  constructor(el: Partial<RefTypeDefinitionType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);
    this.namespace = new DCCXMLElement(el.namespace);
    this.link = new DCCXMLElement(el.link);
    if (el.release) this.release = new DCCXMLElement(el.release);
    if (el.value) this.value = new DCCXMLElement(el.value);
    if (el.procedure) this.procedure = new DCCXMLElement(el.procedure);
  }
}

/* This type contains the information about amended/ substituted DCCs. */
export class ReportAmendedSubstitutedType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  typeOfChange: DCCXMLElement;
  replacedUniqueIdentifier: DCCXMLElement;

  constructor(el: Partial<ReportAmendedSubstitutedType> = {}) {
    super(el);
    this.typeOfChange = new DCCXMLElement(el.typeOfChange);
    this.replacedUniqueIdentifier = new DCCXMLElement(el.replacedUniqueIdentifier);
  }
}

/** List of measurement results that are part of a DCC. */
export class MeasurementResultListType extends DCCXMLElement {
  measurementResult: MeasurementResultType[];

  constructor(el: Partial<MeasurementResultListType> = {}) {
    super(el);
    this.measurementResult = ensureArray(el.measurementResult).map((x) => new MeasurementResultType(x));
  }
}

/** The byteDataType defines a type which allows to add binary encoded files to the measurement result
 * section. */
export class ByteDataType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name?: TextType;
  description?: RichContentType;
  fileName: DCCXMLElement;
  mimeType: DCCXMLElement;
  dataBase64: DCCXMLElement;

  constructor(el: Partial<ByteDataType> = {}) {
    super(el);
    if (el.name) this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);
    this.fileName = new DCCXMLElement(el.fileName);
    this.mimeType = new DCCXMLElement(el.mimeType);
    this.dataBase64 = new DCCXMLElement(el.dataBase64);
  }
}

/** A list of software elements. */
export class SoftwareListType extends DCCXMLElement {
  software: SoftwareType[];

  constructor(el: Partial<SoftwareListType> = {}) {
    super(el);
    this.software = ensureArray(el.software).map((x) => new SoftwareType(x));
  }
}

/** Important metadata for the DCC containing the global unique identifier and other identifications. */
export class CoreDataType extends DCCXMLElement {
  countryCodeISO3166_1: DCCXMLElement;
  usedLangCodeISO639_1: DCCXMLElement[];
  mandatoryLangCodeISO639_1: DCCXMLElement[];
  uniqueIdentifier: DCCXMLElement;
  identifications?: IdentificationListType;
  receiptDate?: DCCXMLDate;
  beginPerformanceDate: DCCXMLDate;
  endPerformanceDate: DCCXMLDate;
  performanceLocation: PerformanceLocationType;
  issueDate?: DCCXMLDate;
  reportAmendedSubstituted?: ReportAmendedSubstitutedType;
  previousReport?: HashType;

  constructor(el: Partial<CoreDataType> = {}) {
    super(el);
    this.countryCodeISO3166_1 = new DCCXMLElement(el.countryCodeISO3166_1);
    this.usedLangCodeISO639_1 = ensureArray(el.usedLangCodeISO639_1).map((x) => new DCCXMLElement(x));
    this.mandatoryLangCodeISO639_1 = ensureArray(el.mandatoryLangCodeISO639_1).map((x) => new DCCXMLElement(x));
    this.uniqueIdentifier = new DCCXMLElement(el.uniqueIdentifier);
    if (el.identifications) this.identifications = new IdentificationListType(el.identifications);
    if (el.receiptDate) this.receiptDate = new DCCXMLDate(el.receiptDate);
    this.beginPerformanceDate = new DCCXMLDate(el.beginPerformanceDate);
    this.endPerformanceDate = new DCCXMLDate(el.endPerformanceDate);
    this.performanceLocation = new PerformanceLocationType(el.performanceLocation);
    if (el.issueDate) this.issueDate = new DCCXMLDate(el.issueDate);
    if (el.reportAmendedSubstituted) this.reportAmendedSubstituted = new ReportAmendedSubstitutedType(el.reportAmendedSubstituted);
    if (el.previousReport) this.previousReport = new HashType(el.previousReport);
  }
}

export class PerformanceLocationType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  constructor(el: Partial<PerformanceLocationType> = {}) {
    super(el);
  }
}

/** This element is a set of calibrated items.
 * Contains one or more item elements. */
export class ItemListType extends DCCXMLElement {
  name?: TextType;
  equipmentClass?: EquipmentClassType[];
  description?: RichContentType;
  owner?: ContactType;
  manufacturer?: ContactNotStrictType;
  identifications?: IdentificationListType;
  item: ItemType[];

  constructor(el: Partial<ItemListType> = {}) {
    super(el);
    if (el.name) this.name = new TextType(el.name);
    if (el.equipmentClass) this.equipmentClass = ensureArray(el.equipmentClass).map((x) => new EquipmentClassType(x));
    if (el.description) this.description = new RichContentType(el.description);
    if (el.owner) this.owner = new ContactType(el.owner);
    if (el.manufacturer) this.manufacturer = new ContactNotStrictType(el.manufacturer);
    if (el.identifications) this.identifications = new IdentificationListType(el.identifications);
    this.item = ensureArray(el.item).map((x) => new ItemType(x));
  }
}

/** Information about the calibration laboratory. */
export class CalibrationLaboratoryType extends DCCXMLElement {
  calibrationLaboratoryCode?: DCCXMLElement;
  contact: ContactType;
  cryptElectronicSeal?: DCCXMLBoolean;
  cryptElectronicSignature?: DCCXMLBoolean;
  cryptElectronicTimeStamp?: DCCXMLBoolean;

  constructor(el: Partial<CalibrationLaboratoryType> = {}) {
    super(el);
    if (el.calibrationLaboratoryCode) this.calibrationLaboratoryCode = new DCCXMLElement(el.calibrationLaboratoryCode);
    this.contact = new ContactType(el.contact);
    if (el.cryptElectronicSeal) this.cryptElectronicSeal = new DCCXMLBoolean(el.cryptElectronicSeal);
    if (el.cryptElectronicSignature) this.cryptElectronicSignature = new DCCXMLBoolean(el.cryptElectronicSignature);
    if (el.cryptElectronicTimeStamp) this.cryptElectronicTimeStamp = new DCCXMLBoolean(el.cryptElectronicTimeStamp);
  }
}

/** List of responsible persons for a DCC. */
export class RespPersonListType extends DCCXMLElement {
  respPerson: RespPersonType[];

  constructor(el: Partial<RespPersonListType> = {}) {
    super(el);
    this.respPerson = ensureArray(el.respPerson).map((x) => new RespPersonType(x));
  }
}

export class ContactType extends DCCXMLElement {
  _attr: IIdAttributes;

  name: TextType;
  eMail?: DCCXMLElement;
  phone?: DCCXMLElement;
  fax?: DCCXMLElement;
  location: LocationType;
  descriptionData?: ByteDataType;

  constructor(el: Partial<ContactType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.eMail) this.eMail = new DCCXMLElement(el.eMail);
    if (el.phone) this.phone = new DCCXMLElement(el.phone);
    if (el.fax) this.fax = new DCCXMLElement(el.fax);
    this.location = new LocationType(el.location);
    if (el.descriptionData) this.descriptionData = new ByteDataType(el.descriptionData);
  }
}

/** List of statements attached to a DCC. */
export class StatementListType extends DCCXMLElement {
  statement: StatementMetaDataType[];

  constructor(el: Partial<StatementListType> = {}) {
    super(el);
    this.statement = ensureArray(el.statement).map((x) => new StatementMetaDataType(x));
  }
}

/** Information about a software including its name, version and a description. */
export class SoftwareType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  name: TextType;
  release: DCCXMLElement;
  type?: DCCXMLElement;
  description?: RichContentType;

  constructor(el: Partial<SoftwareType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    this.release = new DCCXMLElement(el.release);
    if (el.type) this.type = new DCCXMLElement(el.type);
    if (el.description) this.description = new RichContentType(el.description);
  }
}

/** Allow multiple refTypes for one element as attribute. */
export class RefTypesType extends DCCXMLElement {
  constructor(el: Partial<RefTypesType> = {}) {
    super(el);
  }
}

/** The textType defines the type for writing localized text in the DCC.
 * In this Type, the element content can be used many times with different language definition (attribute
 * lang). */
export class TextType extends DCCXMLElement {
  _attr: IIdAttributes;

  content: StringWithLangType[];

  constructor(el: Partial<TextType> = {}) {
    super(el);
    this.content = ensureArray(el.content).map((x) => new StringWithLangType(x));
  }

  getContent(lang?: string): StringWithLangType | undefined {
    // item with undefined as lang is default
    const def = this.content.find((x) => x._attr == null || x._attr.lang == null);

    const items = this.content.filter((x) => x._attr != null && x._attr.lang === lang);

    // if there is more than one content element
    // concat the text of all content elements with same lang
    if (items.length > 1) {
      items.forEach((item) => {
        this.content.splice(this.content.indexOf(item), 1);
      });

      const grouped = items.reduce((prev, curr) => {
        prev._text += `\n${curr._text}`;
        return prev;
      });

      this.content.push(grouped);

      return grouped;
    } else {
      return items.length ? items[0] : def;
    }
  }
}

/** Rich content Type can contain files and formulas beside the normal text content. */
export class RichContentType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name?: TextType;
  content?: StringWithLangType[];
  file?: ByteDataType[];
  formula?: FormulaType[];

  constructor(el: Partial<RichContentType> = {}) {
    super(el);
    if (el.name) this.name = new TextType(el.name);
    if (el.content) this.content = ensureArray(el.content).map((x) => new StringWithLangType(x));
    if (el.file) this.file = ensureArray(el.file).map((x) => new ByteDataType(x));
    if (el.formula) this.formula = ensureArray(el.formula).map((x) => new FormulaType(x));
  }

  getContent(lang?: string): StringWithLangType | undefined {
    if (!this.content) return null;

    // item with undefined as lang is default
    const def = this.content.find((x) => x._attr == null || x._attr.lang == null);

    const items = this.content.filter((x) => x._attr != null && x._attr.lang === lang);

    // if there is more than one content element
    // concat the text of all content elements with same lang
    if (items.length > 1) {
      items.forEach((item) => {
        this.content.splice(this.content.indexOf(item), 1);
      });

      const grouped = items.reduce((prev, curr) => {
        prev._text += `\n${curr._text}`;
        return prev;
      });

      this.content.push(grouped);

      return grouped;
    } else {
      return items.length ? items[0] : def;
    }
  }
}

/** List of measuring equipment and instruments */
export class MeasuringEquipmentListType extends DCCXMLElement {
  name?: TextType;
  equipmentClass?: EquipmentClassType[];
  description?: RichContentType;
  owner?: ContactType;
  identifications?: IdentificationListType;
  measuringEquipment: MeasuringEquipmentType[];

  constructor(el: Partial<MeasuringEquipmentListType> = {}) {
    super(el);
    if (el.name) this.name = new TextType(el.name);
    if (el.equipmentClass) this.equipmentClass = ensureArray(el.equipmentClass).map((x) => new EquipmentClassType(x));
    if (el.description) this.description = new RichContentType(el.description);
    if (el.owner) this.owner = new ContactType(el.owner);
    if (el.identifications) this.identifications = new IdentificationListType(el.identifications);
    this.measuringEquipment = ensureArray(el.measuringEquipment).map((x) => new MeasuringEquipmentType(x));
  }
}

/** Clear name(s) of the item(s) and identifier(s). */
export class EquipmentClassType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  reference: DCCXMLElement;
  classID: DCCXMLElement;

  constructor(el: Partial<EquipmentClassType> = {}) {
    super(el);
    this.reference = new DCCXMLElement(el.reference);
    this.classID = new DCCXMLElement(el.classID);
  }
}

/** List of additional identifications. */
export class IdentificationListType extends DCCXMLElement {
  identification: IdentificationType[];

  constructor(el: Partial<IdentificationListType> = {}) {
    super(el);
    this.identification = ensureArray(el.identification).map((x) => new IdentificationType(x));
  }
}

/** Information about a measuring equipment or instrument used in the calibration */
export class MeasuringEquipmentType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  name: TextType;
  equipmentClass?: EquipmentClassType[];
  description?: RichContentType;
  certificate?: HashType;
  manufacturer?: ContactNotStrictType;
  model?: DCCXMLElement;
  identifications?: IdentificationListType;
  measuringEquipmentQuantities?: MeasuringEquipmentQuantityListType;

  constructor(el: Partial<MeasuringEquipmentType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.equipmentClass) this.equipmentClass = ensureArray(el.equipmentClass).map((x) => new EquipmentClassType(x));
    if (el.description) this.description = new RichContentType(el.description);
    if (el.certificate) this.certificate = new HashType(el.certificate);
    if (el.manufacturer) this.manufacturer = new ContactNotStrictType(el.manufacturer);
    if (el.model) this.model = new DCCXMLElement(el.model);
    if (el.identifications) this.identifications = new IdentificationListType(el.identifications);
    if (el.measuringEquipmentQuantities) this.measuringEquipmentQuantities = new MeasuringEquipmentQuantityListType(el.measuringEquipmentQuantities);
  }
}

export class HashType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  referral: TextType;
  referralID: DCCXMLElement;
  procedure: DCCXMLElement;
  value: DCCXMLElement;
  description?: RichContentType;
  inValidityRange?: DCCXMLBoolean;
  traceable?: DCCXMLBoolean;
  linkedReport?: HashType;

  constructor(el: Partial<HashType> = {}) {
    super(el);
    this.referral = new TextType(el.referral);
    this.referralID = new DCCXMLElement(el.referralID);
    this.procedure = new DCCXMLElement(el.procedure);
    this.value = new DCCXMLElement(el.value);
    if (el.description) this.description = new RichContentType(el.description);
    if (el.inValidityRange) this.inValidityRange = new DCCXMLBoolean(el.inValidityRange);
    if (el.traceable) this.traceable = new DCCXMLBoolean(el.traceable);
    if (el.linkedReport) this.linkedReport = new HashType(el.linkedReport);
  }
}

export class ContactNotStrictType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name: TextType;
  eMail?: DCCXMLElement;
  phone?: DCCXMLElement;
  fax?: DCCXMLElement;
  location?: LocationType;
  descriptionData?: ByteDataType;

  constructor(el: Partial<ContactNotStrictType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.eMail) this.eMail = new DCCXMLElement(el.eMail);
    if (el.phone) this.phone = new DCCXMLElement(el.phone);
    if (el.fax) this.fax = new DCCXMLElement(el.fax);
    if (el.location) this.location = new LocationType(el.location);
    if (el.descriptionData) this.descriptionData = new ByteDataType(el.descriptionData);
  }
}

/** List of values which describes the measurement equipment. */
export class MeasuringEquipmentQuantityListType extends DCCXMLElement {
  measuringEquipmentQuantity: PrimitiveQuantityType[];

  constructor(el: Partial<MeasuringEquipmentQuantityListType> = {}) {
    super(el);
    this.measuringEquipmentQuantity = ensureArray(el.measuringEquipmentQuantity).map((x) => new PrimitiveQuantityType(x));
  }
}

export class PrimitiveQuantityType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name?: TextType;
  description?: RichContentType;

  // region choice
  noQuantity?: RichContentType;
  charsXMLList?: DCCXMLList;
  /** Metadata element definition for a real measurement quantity.
   *
   * The following statements of a real quantity are possible.
   * [(m)-mandatory, (o)-optional]
   *
   * 1. Basic measured quantity
   * (o) - element label        (string)
   * (m) - element value        (decimal value type)
   * (m) - element unit        (string - SI format)
   * (o) - element dateTime      (xs:dateTime)
   *
   * 2. Measured quantity with univariate measurement uncertainty
   * (o) - element label        (string)
   * (m) - element value        (decimal value type)
   * (m) - element unit        (string - SI format)
   * (o) - element dateTime      (xs:dateTime)
   * (m) - choice si:expandedUnc or si:coverageInterval for the univaraite uncertainty
   *
   * Integration into external XML:
   * <myXML xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:si="https://ptb.de/si">
   * <xs:element name="individualElement">
   * <xs:complexType>
   * <xs:sequence>
   * <xs:element ref="si:real"/>
   * </xs:sequence>
   * </xs:complexType>
   * </xs:element>
   * </mxXML> */
  real?: si.RealQuantityType;
  /** The hybrid elements allows to add quantities to the
   * machine readable D-SI format, with other units, than those allowed
   * to be used with the SI by means of the BIPM SI brochure.
   *
   * The hybrid element can contain the following information
   *
   * 1. A set of real quantities
   * - all real elements provide a quantity value for one and the same measured quantity
   * - each real element provides this quantity with a different unit
   * - at least one real element provides the quantity with a machine readable SI unit
   * - the other real quantities can use any SI or non-SI unit
   *
   * 2. A set of complex quantities
   * - all complex elements provide a quantity value for one and the same measured quantity
   * - each complex element provides this quantity with a different unit(s)
   * - at least one complex element provides the quantity with a machine readable SI unit(s)
   * - the other complex quantities can use any SI or non-SI unit
   *
   * 3. A set of list elements
   * - all list elements must provide the same quantity information and hence, must have
   * an identical structure.
   * - the lists do only differ by using different units for each of the quantities
   * - at least one list  provides all quantities only with machine readable SI units
   * - the other lists can use any other units for the quantities
   * - Applicable list types are si:list, si:realList, si:complexList and si:realListXMLList.
   *
   * 4. A set of constant quantities
   * - all constant elements provide a quantity value for one and the same quantity
   * - each constant element provides this quantity with a different unit
   * - at least one constant element provides the quantity with a machine readable SI unit
   * - the other constant quantities can use any SI or non-SI unit */
  hybrid?: si.HybridType;
  /** The definition of complex quantities in the D-SI metadata model.
   * Complex quantities allow two representations of complex numerical values:
   * One is the Cartesian coordinate form, the other representation is the
   * polar coordinate form.
   *
   * The following statements of a complex quantity are possible.
   * [(m)-mandatory, (o)-optional]
   *
   * 1. Basic measured quantity in Cartesian form
   * (o) - element label        (string)
   * (m) - element valueReal      (decimal value type)
   * (m) - element valueImag      (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (o) - element dateTime      (xs:dateTime)
   *
   * 2. Basic measured quantity in polar form
   * (o) - element label        (string)
   * (m) - element valueMagnitude  (decimal value type)
   * (m) - element valuePhase    (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (m) - element unitPhase      (string - SI unit for an angular quantity)
   * (o) - element dateTime      (xs:dateTime)
   *
   * 3. Basic measured quantity in Cartesian form with bivariate uncertainty
   * (o) - element label        (string)
   * (m) - element valueReal      (decimal value type)
   * (m) - element valueImag      (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (o) - element dateTime      (xs:dateTime)
   * (m) - choice si:ellipsoidalRegion or si:rectangularRegion for the bivariate uncertainty
   *
   * 4. Basic measured quantity in polar form with bivariate uncertainty
   * (o) - element label        (string)
   * (m) - element valueMagnitude  (decimal value type)
   * (m) - element valuePhase    (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (m) - element unitPhase      (string - SI unit for an angular quantity)
   * (o) - element dateTime      (xs:dateTime)
   * (m) - choice si:ellipsoidalRegion or si:rectangularRegion for the bivariate uncertainty */
  complex?: si.ComplexQuantityType;
  /** Definition of a structure for real numbers, that represent for fundamental
   * physical constants and mathematical  constants.
   *
   * The element has the following components [(m)-mandatory, (o)-optional]:
   *
   * (o) - element label            (string)
   * (m) - element value            (decimal value type)
   * (m) - element unit            (string - SI unit)
   * (o) - element dateTime          (xs:dateTime)
   * (o) - element uncertainty        (decimal value >= 0)
   * (o) - element distribution        (string)
   *
   * The value and the uncertainty have the unit specified by the element unit.
   * For fundamental physical constants, that are defined experimentally, the
   * uncertainty is the standard deviation.
   *
   * For rounded mathematical constants, the uncertainty is the standard deviation
   * of a rectangular distribution (element value defines center point), that
   * contains the exact value of the constant with 100 percent probability. */
  constant?: si.ConstantQuantityType;
  /** Meta data element definition for a list of real measurement quantities based on
   * the XML type xsd:list. The list can represent independent measurement or multivariate
   * vector quantities. The listed data is separated by blank spaces allowing a memory
   * saving implementation of large data sets.
   *
   * A list can provide the following structures:
   * - list of one ore more si:value elements separated by blank spaces
   * - optional list timestamp, list label, and/or list unit
   * - optional list univariate uncertainty statement with list unit
   * - optional multivariate hyper-elliptical or hyper-rectangular coverage region */
  realListXMLList?: si.RealListXMLListType;

  // endregion choice

  constructor(el: Partial<PrimitiveQuantityType> = {}) {
    super(el);
    if (el.name) this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);

    // choice
    if (el.noQuantity) this.noQuantity = new RichContentType(el.noQuantity);
    else if (el.charsXMLList) this.charsXMLList = new DCCXMLList(el.charsXMLList);
    else if (el.real) this.real = new si.RealQuantityType(el.real);
    else if (el.hybrid) this.hybrid = new si.HybridType(el.hybrid);
    else if (el.complex) this.complex = new si.ComplexQuantityType(el.complex);
    else if (el.constant) this.constant = new si.ConstantQuantityType(el.constant);
    else if (el.realListXMLList) this.realListXMLList = new si.RealListXMLListType(el.realListXMLList);
  }
}

/** A item that is calibrated in this DCC. */
export class ItemType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  name: TextType;
  equipmentClass?: EquipmentClassType[];
  description?: RichContentType;
  installedSoftwares?: SoftwareListType;
  manufacturer?: ContactNotStrictType;
  model?: DCCXMLElement;
  identifications: IdentificationListType;
  itemQuantities: ItemQuantityListType;

  constructor(el: Partial<ItemType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.equipmentClass) this.equipmentClass = ensureArray(el.equipmentClass).map((x) => new EquipmentClassType(x));
    if (el.description) this.description = new RichContentType(el.description);
    if (el.installedSoftwares) this.installedSoftwares = new SoftwareListType(el.installedSoftwares);
    if (el.manufacturer) this.manufacturer = new ContactNotStrictType(el.manufacturer);
    if (el.model) this.model = new DCCXMLElement(el.model);
    this.identifications = new IdentificationListType(el.identifications);
    if (el.itemQuantities) this.itemQuantities = new ItemQuantityListType(el.itemQuantities);
  }
}

/** List of value(s) which belongs to the to be calibrated item. */
export class ItemQuantityListType extends DCCXMLElement {
  itemQuantity: PrimitiveQuantityType[];

  constructor(el: Partial<ItemQuantityListType> = {}) {
    super(el);
    this.itemQuantity = ensureArray(el.itemQuantity).map((x) => new PrimitiveQuantityType(x));
  }
}

/** An additional identification (eg. reference no., serial number, etc.). */
export class IdentificationType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  issuer: DCCXMLElement;
  value: DCCXMLElement;
  name?: TextType;

  constructor(el: Partial<IdentificationType> = {}) {
    super(el);
    this.issuer = new DCCXMLElement(el.issuer);
    this.value = new DCCXMLElement(el.value);
    if (el.name) this.name = new TextType(el.name);
  }
}

/** A person responsible for a DCC. */
export class RespPersonType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  person: ContactNotStrictType;
  description?: RichContentType;
  role?: DCCXMLElement;
  mainSigner?: DCCXMLBoolean;
  cryptElectronicSeal?: DCCXMLBoolean;
  cryptElectronicSignature?: DCCXMLBoolean;
  cryptElectronicTimeStamp?: DCCXMLBoolean;

  constructor(el: Partial<RespPersonType> = {}) {
    super(el);
    this.person = new ContactNotStrictType(el.person);
    if (el.description) this.description = new RichContentType(el.description);
    if (el.role) this.role = new DCCXMLElement(el.role);
    if (el.mainSigner) this.mainSigner = new DCCXMLBoolean(el.mainSigner);
    if (el.cryptElectronicSeal) this.cryptElectronicSeal = new DCCXMLBoolean(el.cryptElectronicSeal);
    if (el.cryptElectronicSignature) this.cryptElectronicSignature = new DCCXMLBoolean(el.cryptElectronicSignature);
    if (el.cryptElectronicTimeStamp) this.cryptElectronicTimeStamp = new DCCXMLBoolean(el.cryptElectronicTimeStamp);
  }
}

export class StatementMetaDataType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name?: TextType;
  description?: RichContentType;
  countryCodeISO3166_1?: DCCXMLElement[];
  convention?: DCCXMLElement;
  traceable?: DCCXMLBoolean;
  norm?: DCCXMLElement[];
  reference?: DCCXMLElement[];
  declaration?: RichContentType;

  // region choice
  valid?: DCCXMLBoolean;
  validXMLList?: DCCXMLList;
  // endregion choice

  date?: DCCXMLDate;
  period?: DCCXMLElement;
  respAuthority?: ContactType;

  // region choice
  conformity?: DCCXMLElement;
  conformityXMLList?: DCCXMLList;
  // endregion choice

  data?: DataType;
  nonSIDefinition?: DCCXMLElement;
  nonSIUnit?: DCCXMLElement;
  location?: LocationType;

  constructor(el: Partial<StatementMetaDataType> = {}) {
    super(el);
    if (el.name) this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);
    if (el.countryCodeISO3166_1) this.countryCodeISO3166_1 = ensureArray(el.countryCodeISO3166_1).map((x) => new DCCXMLElement(x));
    if (el.convention) this.convention = new DCCXMLElement(el.convention);
    if (el.traceable) this.traceable = new DCCXMLBoolean(el.traceable);
    if (el.norm) this.norm = ensureArray(el.norm).map((x) => new DCCXMLElement(x));
    if (el.reference) this.reference = ensureArray(el.reference).map((x) => new DCCXMLElement(x));
    if (el.declaration) this.declaration = new RichContentType(el.declaration);

    // choice
    if (el.valid) this.valid = new DCCXMLBoolean(el.valid);
    else if (el.validXMLList) this.validXMLList = new DCCXMLList(el.validXMLList);

    if (el.date) this.date = new DCCXMLDate(el.date);
    if (el.period) this.period = new DCCXMLElement(el.period);
    if (el.respAuthority) this.respAuthority = new ContactType(el.respAuthority);

    // choice
    if (el.conformity) this.conformity = new DCCXMLElement(el.conformity);
    else if (el.conformityXMLList) this.conformityXMLList = new DCCXMLList(el.conformityXMLList);

    if (el.data) this.data = new DataType(el.data);
    if (el.nonSIDefinition) this.nonSIDefinition = new DCCXMLElement(el.nonSIDefinition);
    if (el.nonSIUnit) this.nonSIUnit = new DCCXMLElement(el.nonSIUnit);
    if (el.location) this.location = new LocationType(el.location);
  }
}

/** A measurement results with the methods, software and equipments used for the calibration.
 * Also contains influence conditions and a list of the actual results. */
export class MeasurementResultType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name: TextType;
  description?: RichContentType;
  usedMethods?: UsedMethodListType;
  usedSoftware?: SoftwareListType;
  measuringEquipments?: MeasuringEquipmentListType;
  influenceConditions?: InfluenceConditionListType;
  results: ResultListType;
  measurementMetaData?: MeasurementMetaDataListType;

  constructor(el: Partial<MeasurementResultType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);
    if (el.usedMethods) this.usedMethods = new UsedMethodListType(el.usedMethods);
    if (el.usedSoftware) this.usedSoftware = new SoftwareListType(el.usedSoftware);
    if (el.measuringEquipments) this.measuringEquipments = new MeasuringEquipmentListType(el.measuringEquipments);
    if (el.influenceConditions) this.influenceConditions = new InfluenceConditionListType(el.influenceConditions);
    this.results = new ResultListType(el.results);
    if (el.measurementMetaData) this.measurementMetaData = new MeasurementMetaDataListType(el.measurementMetaData);
  }
}

/** List of the methods used in the calibration. */
export class UsedMethodListType extends DCCXMLElement {
  usedMethod: UsedMethodType[];

  constructor(el: Partial<UsedMethodListType> = {}) {
    super(el);
    this.usedMethod = ensureArray(el.usedMethod).map((x) => new UsedMethodType(x));
  }
}

/** List of conditions that influence the measurements. */
export class InfluenceConditionListType extends DCCXMLElement {
  influenceCondition: ConditionType[];

  constructor(el: Partial<InfluenceConditionListType> = {}) {
    super(el);
    this.influenceCondition = ensureArray(el.influenceCondition).map((x) => new ConditionType(x));
  }
}

/** List of results of the calibration. */
export class ResultListType extends DCCXMLElement {
  result: ResultType[];

  constructor(el: Partial<ResultListType> = {}) {
    super(el);
    this.result = ensureArray(el.result).map((x) => new ResultType(x));
  }
}

/** A list of additional metadata elements. */
export class MeasurementMetaDataListType extends DCCXMLElement {
  metaData: StatementMetaDataType[];

  constructor(el: Partial<MeasurementMetaDataListType> = {}) {
    super(el);
    this.metaData = ensureArray(el.metaData).map((x) => new StatementMetaDataType(x));
  }
}

/** A method used in the calibration. */
export class UsedMethodType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  name: TextType;
  description?: RichContentType;
  norm?: DCCXMLElement[];
  reference?: DCCXMLElement[];

  constructor(el: Partial<UsedMethodType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);
    if (el.norm) this.norm = ensureArray(el.norm).map((x) => new DCCXMLElement(x));
    if (el.reference) this.reference = ensureArray(el.reference).map((x) => new DCCXMLElement(x));
  }
}

/** Condition (e.g. environmental) under which the calibrations were performed which have an influence on
 * the measurement results. */
export class ConditionType extends DCCXMLElement {
  _attr: IIdAndRefTypeAttributes;

  name: TextType;
  description?: RichContentType;
  status?: DCCXMLElement;
  certificate?: HashType;
  data: DataType;

  constructor(el: Partial<ConditionType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);
    if (el.status) this.status = new DCCXMLElement(el.status);
    if (el.certificate) this.certificate = new HashType(el.certificate);
    this.data = new DataType(el.data);
  }
}

export class DataType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  // region choice
  text?: RichContentType[];
  formula?: FormulaType[];
  byteData?: ByteDataType[];
  xml?: XmlType[];
  quantity?: QuantityType[];
  list?: ListType[];

  // endregion choice

  constructor(el: Partial<DataType> = {}) {
    super(el);

    // choice but maxOccurs = unbounded
    if (el.text) this.text = ensureArray(el.text).map((x) => new RichContentType(x));
    if (el.formula) this.formula = ensureArray(el.formula).map((x) => new FormulaType(x));
    if (el.byteData) this.byteData = ensureArray(el.byteData).map((x) => new ByteDataType(x));
    if (el.xml) this.xml = ensureArray(el.xml).map((x) => new XmlType(x));
    if (el.quantity) this.quantity = ensureArray(el.quantity).map((x) => new QuantityType(x));
    if (el.list) this.list = ensureArray(el.list).map((x) => new ListType(x));
  }
}

/** The actual result of the calibration. */
export class ResultType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name: TextType;
  description?: RichContentType;
  data: DataType;

  constructor(el: Partial<ResultType> = {}) {
    super(el);
    this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);
    this.data = new DataType(el.data);
  }
}

/** This data block is used to add formulas and equations to the measurement result section of the DCC. */
export class FormulaType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  // region choice
  latex?: DCCXMLElement;
  mathml?: XmlType;

  // endregion choice

  constructor(el: Partial<FormulaType> = {}) {
    super(el);

    // choice
    if (el.latex) this.latex = new DCCXMLElement(el.latex);
    else if (el.mathml) this.mathml = new XmlType(el.mathml);
  }
}

/** This data block is used to add user or application specific XML content. */
export class XmlType extends DCCXMLElement {
  [key: string]: unknown;
  _attr: IIdAndRefIdAndRefTypeAttributes;

  constructor(el: Partial<XmlType> = {}) {
    super(el);
    Object.assign(this, el);
  }
}

/** A single measurement quantity that can contain a D-SI element or text.
 * Additionally used methods, used software, influence conditions that only affect this quantity can be
 * added.
 * Measurement metadata can also be added. */
export class QuantityType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name?: TextType;
  description?: RichContentType;

  // region choice
  noQuantity?: RichContentType;
  charsXMLList?: DCCXMLList;
  /** Metadata element definition for a real measurement quantity.
   *
   * The following statements of a real quantity are possible.
   * [(m)-mandatory, (o)-optional]
   *
   * 1. Basic measured quantity
   * (o) - element label        (string)
   * (m) - element value        (decimal value type)
   * (m) - element unit        (string - SI format)
   * (o) - element dateTime      (xs:dateTime)
   *
   * 2. Measured quantity with univariate measurement uncertainty
   * (o) - element label        (string)
   * (m) - element value        (decimal value type)
   * (m) - element unit        (string - SI format)
   * (o) - element dateTime      (xs:dateTime)
   * (m) - choice si:expandedUnc or si:coverageInterval for the univaraite uncertainty
   *
   * Integration into external XML:
   * <myXML xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:si="https://ptb.de/si">
   * <xs:element name="individualElement">
   * <xs:complexType>
   * <xs:sequence>
   * <xs:element ref="si:real"/>
   * </xs:sequence>
   * </xs:complexType>
   * </xs:element>
   * </mxXML> */
  real?: si.RealQuantityType;
  /** Meta data element definition for a list of basic measurement quantities.
   * The list can represent independent measurement or multivariate vector quantities.
   *
   * This list can provide the following nested structures:
   * - si:list elements and
   * - si:realList elements and
   * - si:complexList elements */
  list?: si.SIListType;
  /** The hybrid elements allows to add quantities to the
   * machine readable D-SI format, with other units, than those allowed
   * to be used with the SI by means of the BIPM SI brochure.
   *
   * The hybrid element can contain the following information
   *
   * 1. A set of real quantities
   * - all real elements provide a quantity value for one and the same measured quantity
   * - each real element provides this quantity with a different unit
   * - at least one real element provides the quantity with a machine readable SI unit
   * - the other real quantities can use any SI or non-SI unit
   *
   * 2. A set of complex quantities
   * - all complex elements provide a quantity value for one and the same measured quantity
   * - each complex element provides this quantity with a different unit(s)
   * - at least one complex element provides the quantity with a machine readable SI unit(s)
   * - the other complex quantities can use any SI or non-SI unit
   *
   * 3. A set of list elements
   * - all list elements must provide the same quantity information and hence, must have
   * an identical structure.
   * - the lists do only differ by using different units for each of the quantities
   * - at least one list  provides all quantities only with machine readable SI units
   * - the other lists can use any other units for the quantities
   * - Applicable list types are si:list, si:realList, si:complexList and si:realListXMLList.
   *
   * 4. A set of constant quantities
   * - all constant elements provide a quantity value for one and the same quantity
   * - each constant element provides this quantity with a different unit
   * - at least one constant element provides the quantity with a machine readable SI unit
   * - the other constant quantities can use any SI or non-SI unit */
  hybrid?: si.HybridType;
  /** The definition of complex quantities in the D-SI metadata model.
   * Complex quantities allow two representations of complex numerical values:
   * One is the Cartesian coordinate form, the other representation is the
   * polar coordinate form.
   *
   * The following statements of a complex quantity are possible.
   * [(m)-mandatory, (o)-optional]
   *
   * 1. Basic measured quantity in Cartesian form
   * (o) - element label        (string)
   * (m) - element valueReal      (decimal value type)
   * (m) - element valueImag      (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (o) - element dateTime      (xs:dateTime)
   *
   * 2. Basic measured quantity in polar form
   * (o) - element label        (string)
   * (m) - element valueMagnitude  (decimal value type)
   * (m) - element valuePhase    (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (m) - element unitPhase      (string - SI unit for an angular quantity)
   * (o) - element dateTime      (xs:dateTime)
   *
   * 3. Basic measured quantity in Cartesian form with bivariate uncertainty
   * (o) - element label        (string)
   * (m) - element valueReal      (decimal value type)
   * (m) - element valueImag      (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (o) - element dateTime      (xs:dateTime)
   * (m) - choice si:ellipsoidalRegion or si:rectangularRegion for the bivariate uncertainty
   *
   * 4. Basic measured quantity in polar form with bivariate uncertainty
   * (o) - element label        (string)
   * (m) - element valueMagnitude  (decimal value type)
   * (m) - element valuePhase    (decimal value type)
   * (m) - element unit        (string - SI unit)
   * (m) - element unitPhase      (string - SI unit for an angular quantity)
   * (o) - element dateTime      (xs:dateTime)
   * (m) - choice si:ellipsoidalRegion or si:rectangularRegion for the bivariate uncertainty */
  complex?: si.ComplexQuantityType;
  /** Definition of a structure for real numbers, that represent for fundamental
   * physical constants and mathematical  constants.
   *
   * The element has the following components [(m)-mandatory, (o)-optional]:
   *
   * (o) - element label            (string)
   * (m) - element value            (decimal value type)
   * (m) - element unit            (string - SI unit)
   * (o) - element dateTime          (xs:dateTime)
   * (o) - element uncertainty        (decimal value >= 0)
   * (o) - element distribution        (string)
   *
   * The value and the uncertainty have the unit specified by the element unit.
   * For fundamental physical constants, that are defined experimentally, the
   * uncertainty is the standard deviation.
   *
   * For rounded mathematical constants, the uncertainty is the standard deviation
   * of a rectangular distribution (element value defines center point), that
   * contains the exact value of the constant with 100 percent probability. */
  constant?: si.ConstantQuantityType;
  /** Meta data element definition for a list of real measurement quantities based on
   * the XML type xsd:list. The list can represent independent measurement or multivariate
   * vector quantities. The listed data is separated by blank spaces allowing a memory
   * saving implementation of large data sets.
   *
   * A list can provide the following structures:
   * - list of one ore more si:value elements separated by blank spaces
   * - optional list timestamp, list label, and/or list unit
   * - optional list univariate uncertainty statement with list unit
   * - optional multivariate hyper-elliptical or hyper-rectangular coverage region */
  realListXMLList?: si.RealListXMLListType;
  // endregion choice

  relativeUncertainty?: RelativeUncertaintyType;
  usedMethods?: UsedMethodListType;
  usedSoftware?: SoftwareListType;
  measuringEquipments?: MeasuringEquipmentListType;
  influenceConditions?: InfluenceConditionListType;
  measurementMetaData?: MeasurementMetaDataListType;

  constructor(el: Partial<QuantityType> = {}) {
    super(el);
    if (el.name) this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);

    // choice
    if (el.noQuantity) this.noQuantity = new RichContentType(el.noQuantity);
    else if (el.charsXMLList) this.charsXMLList = new DCCXMLList(el.charsXMLList);
    else if (el.real) this.real = new si.RealQuantityType(el.real);
    else if (el.list) this.list = new si.SIListType(el.list);
    else if (el.hybrid) this.hybrid = new si.HybridType(el.hybrid);
    else if (el.complex) this.complex = new si.ComplexQuantityType(el.complex);
    else if (el.constant) this.constant = new si.ConstantQuantityType(el.constant);
    else if (el.realListXMLList) this.realListXMLList = new si.RealListXMLListType(el.realListXMLList);

    if (el.relativeUncertainty) this.relativeUncertainty = new RelativeUncertaintyType(el.relativeUncertainty);
    if (el.usedMethods) this.usedMethods = new UsedMethodListType(el.usedMethods);
    if (el.usedSoftware) this.usedSoftware = new SoftwareListType(el.usedSoftware);
    if (el.measuringEquipments) this.measuringEquipments = new MeasuringEquipmentListType(el.measuringEquipments);
    if (el.influenceConditions) this.influenceConditions = new InfluenceConditionListType(el.influenceConditions);
    if (el.measurementMetaData) this.measurementMetaData = new MeasurementMetaDataListType(el.measurementMetaData);
  }
}

/** The 'list' element allows to define a collection of measurement results
 * which are subject to structures with integrity. Basic structures are
 * vector quantities. A recursive usage of 'list' allows the creation of
 * matrix and tensor structures as well as structures of higher dimension. The 'list'
 * may also be used to give measurement results in combination with some
 * ambient conditions at the measurement. */
export class ListType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  name?: TextType;
  description?: RichContentType;
  usedMethods?: UsedMethodListType;
  usedSoftware?: SoftwareListType;
  measuringEquipments?: MeasuringEquipmentListType;
  influenceConditions?: InfluenceConditionListType;
  measurementMetaData?: MeasurementMetaDataListType;

  // region choice
  dateTime?: DCCXMLDate;
  dateTimeXMLList?: DCCXMLList;
  // endregion choice

  // region choice
  list?: ListType[];
  quantity?: QuantityType[];
  // endregion choice

  constructor(el: Partial<ListType> = {}) {
    super(el);

    if (el.name) this.name = new TextType(el.name);
    if (el.description) this.description = new RichContentType(el.description);

    // choice
    if (el.dateTime) this.dateTime = new DCCXMLDate(el.dateTime);
    else if (el.dateTimeXMLList) this.dateTimeXMLList = new DCCXMLList(el.dateTimeXMLList);

    // choice but maxOccurs = unbounded
    if (el.list) this.list = ensureArray(el.list).map((x) => new ListType(x));
    if (el.quantity) this.quantity = ensureArray(el.quantity).map((x) => new QuantityType(x));

    if (el.usedMethods) this.usedMethods = new UsedMethodListType(el.usedMethods);
    if (el.usedSoftware) this.usedSoftware = new SoftwareListType(el.usedSoftware);
    if (el.measuringEquipments) this.measuringEquipments = new MeasuringEquipmentListType(el.measuringEquipments);
    if (el.influenceConditions) this.influenceConditions = new InfluenceConditionListType(el.influenceConditions);
    if (el.measurementMetaData) this.measurementMetaData = new MeasurementMetaDataListType(el.measurementMetaData);
  }
}

export class RelativeUncertaintyType extends DCCXMLElement {
  // region choice
  relativeUncertaintyXmlList?: RelativeUncertaintyXmlList;
  relativeUncertaintySingle?: RelativeUncertaintySingle;
  // endregion choice

  constructor(el: Partial<RelativeUncertaintyType> = {}) {
    super(el);

    // choice
    if (el.relativeUncertaintyXmlList) this.relativeUncertaintyXmlList = new RelativeUncertaintyXmlList(el.relativeUncertaintyXmlList);
    else if (el.relativeUncertaintySingle) this.relativeUncertaintySingle = new RelativeUncertaintySingle(el.relativeUncertaintySingle);
  }
}

export class RelativeUncertaintyXmlList extends si.RealListXMLListType {
  get ns(): string {
    return DCC.namespace;
  }
}

export class RelativeUncertaintySingle extends si.RealQuantityType {
  get ns(): string {
    return DCC.namespace;
  }
}

export interface IStringWithLangTypeAttributes extends IIdAndRefIdAndRefTypeAttributes {
  lang?: string;
}

/** A string element with an additional lang attribute for localization. */
export class StringWithLangType extends DCCXMLElement {
  _attr: IStringWithLangTypeAttributes;

  constructor(el: Partial<StringWithLangType> = {}) {
    super(el);
  }
}

export class LocationType extends DCCXMLElement {
  city: DCCXMLElement[];
  countryCode: DCCXMLElement[];
  postCode: DCCXMLElement[];
  postOfficeBox: DCCXMLElement[];
  state: DCCXMLElement[];
  street: DCCXMLElement[];
  streetNo: DCCXMLElement[];
  further: RichContentType[];
  positionCoordinates: PositionCoordinatesType[];

  constructor(el: Partial<LocationType> = {}) {
    super(el);
    this.city = ensureArray(el.city).map((x) => new DCCXMLElement(x));
    this.countryCode = ensureArray(el.countryCode).map((x) => new DCCXMLElement(x));
    this.postCode = ensureArray(el.postCode).map((x) => new DCCXMLElement(x));
    this.postOfficeBox = ensureArray(el.postOfficeBox).map((x) => new DCCXMLElement(x));
    this.state = ensureArray(el.state).map((x) => new DCCXMLElement(x));
    this.street = ensureArray(el.street).map((x) => new DCCXMLElement(x));
    this.streetNo = ensureArray(el.streetNo).map((x) => new DCCXMLElement(x));
    this.further = ensureArray(el.further).map((x) => new RichContentType(x));
    this.positionCoordinates = ensureArray(el.positionCoordinates).map((x) => new PositionCoordinatesType(x));
  }
}

export class PositionCoordinatesType extends DCCXMLElement {
  _attr: IIdAndRefIdAndRefTypeAttributes;

  positionCoordinateSystem: DCCXMLElement;
  reference?: DCCXMLElement;
  declaration?: RichContentType;
  positionCoordinate1: si.RealQuantityType;
  positionCoordinate2: si.RealQuantityType;
  positionCoordinate3?: si.RealQuantityType;

  constructor(el: Partial<PositionCoordinatesType> = {}) {
    super(el);
    this.positionCoordinateSystem = new DCCXMLElement(el.positionCoordinateSystem);
    if (el.reference) this.reference = new DCCXMLElement(el.reference);
    if (el.declaration) this.declaration = new RichContentType(el.declaration);
    this.positionCoordinate1 = new si.RealQuantityType(el.positionCoordinate1);
    this.positionCoordinate2 = new si.RealQuantityType(el.positionCoordinate2);
    if (el.positionCoordinate3) this.positionCoordinate3 = new si.RealQuantityType(el.positionCoordinate3);
  }
}

export class Comment extends DCCXMLElement {
  constructor(el: Partial<Comment | never>) {
    super(el);
    Object.assign(this, el);
  }
}
